# -*- coding: utf-8 -*-
# Copyright (C) 2020 Adrien Delle Cave
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.classes.abstract.modules"""

import abc
import logging
import os
import threading

from dwho.classes.errors import DWhoConfigurationError
from dwho.classes.modules import DWhoModuleBase
from sonicprobe.libs import anysql, mysql_config_parser, urisup

from ..lock_db import BackupsLockDB


LOG           = logging.getLogger('xtrargon_ng.classes.abstract.modules')

DB_URI_PARAMS = (('autocommit', '1'),
                 ('charset', 'utf8'),
                 ('cursorclass', 'Dict'),
                 ('wait_timeout', '2147483'))


class XtrargonDBModuleBase(DWhoModuleBase):
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def MODULE_NAME(self):
        return

    def __init__(self):
        super(XtrargonDBModuleBase, self).__init__()
        self._db_lock  = None

        self.lock_db   = None
        self.db_locked = None

        self.dbconf    = None
        self.dburi     = None

    def _build_mysql_uri(self):
        host = mysql_config_parser.MYSQL_DEFAULT_HOST
        port = mysql_config_parser.MYSQL_DEFAULT_PORT

        if self.dbconf.has_option('client', 'host'):
            host = self.dbconf.get('client', 'host')

        if self.dbconf.has_option('client', 'port'):
            port = self.dbconf.get('client', 'port')

        uri = ('mysql',
               (self.dbconf.get('client', 'user'),
                self.dbconf.get('client', 'password'),
                host,
                port),
               None,
               DB_URI_PARAMS,
               None)

        return urisup.uri_help_unsplit(uri)

    @staticmethod
    def _check_db_server(uri):
        dbconn = None
        try:
            dbconn = anysql.connect_by_uri(uri)
            if not dbconn.is_connected():
                raise DWhoConfigurationError("unable to connect to database")
        finally:
            if dbconn:
                dbconn.close()

    # pylint: disable-msg=attribute-defined-outside-init
    def _init_db(self):
        self.lock_db   = threading.Event()
        self.db_locked = threading.Event()

        self.dbconf    = mysql_config_parser.MySQLConfigVersion(
            custom_file = os.path.join('client.d', "%s.cnf" % self.server_id),
            config_dir  = self.config.get('_config_directory')).get_client()
        self.dburi     = self._build_mysql_uri()

    def _init_lock_db(self, config):
        if self.dbconf and self.dburi:
            self._check_db_server(self.dburi)
            if self._db_lock:
                self._terminate_lock_db()
            self._db_lock = BackupsLockDB(self, config)
            self._db_lock.daemon = True

    def _start_lock_db(self):
        if self._db_lock:
            self._db_lock.start()

    def _terminate_lock_db(self):
        if self._db_lock:
            self._db_lock.terminate()
            self._db_lock = None

    def _clear_lock_db(self):
        if self.lock_db:
            self.lock_db.clear()
        if self.db_locked:
            self.db_locked.clear()
