# -*- coding: utf-8 -*-
# Copyright (C) 2019 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.classes.s3"""

import logging
import time
import os
import threading

from email.parser import Parser

import bitmath

import botocore
try:
    from botocore.hooks import AliasedEventEmitter as EventEmitter
except ImportError:
    from botocore.hooks import HierarchicalEmitter as EventEmitter

from awscli import clidriver, EnvironmentVariables
from boto3 import session
from boto3.s3 import transfer as S3Transfer

from dwho.classes.errors import DWhoConfigurationError
from sonicprobe import helpers
from sonicprobe.libs import urisup

import six
try:
    from six.moves import cStringIO as StringIO
except ImportError:
    from six import StringIO

import pycurl

LOG = logging.getLogger('xtrargon_ng.classes.s3')

XTRARGON_FILE         = ".xtrargon"

DEFAULT_HTTP_HEADERS  = ('Connection: keep-alive',
                         'Content-Type: application/octet-stream')
DEFAULT_MAX_REDIRS    = 5
DEFAULT_MAX_RETRIES   = 0
DEFAULT_TIMEOUT       = 60
DEFAULT_RETRY_DELAY   = 0.0
DEFAULT_STORAGE_CLASS = 'STANDARD'

_CONFIG_OPTIONS       = ('connect_timeout',
                         'read_timeout')


class XtrargonBotoNullEventHooks(botocore.hooks.BaseEventHooks):
    def _register(self, event_name, handler, unique_id=None, unique_id_uses_count=False): # pylint: disable-msg=unused-argument,no-self-use
        return

    def emit_until_response(self, event_name, **kwargs): # pylint: disable-msg=unused-argument,no-self-use
        return (None, None)


class XtrargonS3Helper(object): # pylint: disable-msg=useless-object-inheritance
    def __init__(self, config = None, client_cfg = None, transfer_cfg = None):
        if not config or not isinstance(config, dict):
            config = {}

        self._config         = config
        self._client_cfg     = None
        self._transfer_cfg   = None
        self._botocore_sess  = None
        self._boto3_sess     = None
        self.connect_timeout = None
        self.max_redirs      = None
        self.max_retries     = None
        self.parse_headers   = None
        self.read_timeout    = None
        self.retry_delay     = None
        self.retry_factor    = None

        if 'client' in config and not client_cfg:
            self._client_cfg = botocore.client.Config(**config['client'])

        if 'transfer' in config and not transfer_cfg:
            self._transfer_cfg = S3Transfer.TransferConfig(**config['transfer'])

        self.set_options(config, from_boto = True)

    @staticmethod
    def create_driver():
        emitter  = EventEmitter()
        bsession = botocore.session.Session(EnvironmentVariables, event_hooks = emitter)
        clidriver._set_user_agent_for_session(bsession) # pylint: disable-msg=protected-access
        clidriver.load_plugins(bsession.full_config.get('plugins', {}),
                               event_hooks = emitter)
        return clidriver.CLIDriver(session = bsession).main

    def botocore_session(self, event_hooks = None):
        if self._botocore_sess:
            return self._botocore_sess

        self._botocore_sess = botocore.session.Session(include_builtin_handlers = False,
                                                       event_hooks = event_hooks)

        return self._botocore_sess

    def boto3_session(self, event_hooks = None):
        if self._boto3_sess:
            return self._boto3_sess

        self._boto3_sess = session.Session(
            botocore_session = self.botocore_session(event_hooks = event_hooks))

        return self._boto3_sess

    def botocore_client(self, service = 's3', event_hooks = None):
        sess = self.botocore_session(event_hooks = event_hooks)

        while True:
            try:
                return sess.create_client(service, config = self._client_cfg)
            except KeyError:
                time.sleep(0.2)

    def boto3_client(self, service = 's3', event_hooks = None):
        sess = self.boto3_session(
            self.botocore_session(event_hooks = event_hooks))

        while True:
            try:
                return sess.client(service, config = self._client_cfg)
            except KeyError:
                time.sleep(0.2)

    def client_config(self):
        return self._client_cfg

    def transfer_config(self):
        return self._transfer_cfg

    def load_credentials(self, credentials = None):
        if not credentials and 'credentials' in self._config:
            credentials = self._config['credentials']

        cred = helpers.load_yaml_file(credentials)
        if not cred:
            raise DWhoConfigurationError("unable to load credentials")

        for k, v in six.iteritems(cred['s3']):
            os.environ[k.upper()] = v

    def get_bucket(self):
        if os.environ.get('AWS_DEFAULT_BUCKET'):
            return os.environ['AWS_DEFAULT_BUCKET']

        return self._config.get('bucket')

    def get_storage_class(self):
        if os.environ.get('AWS_STORAGE_CLASS'):
            return os.environ['AWS_STORAGE_CLASS']

        return self._config.get('storage_class') or DEFAULT_STORAGE_CLASS

    def get_tags(self):
        tags = None
        r    = []

        if os.environ.get('AWS_S3_TAGS'):
            tags = dict(urisup.split_query(os.environ.get('AWS_S3_TAGS')))

        if self._config.get('tags') and isinstance(self._config['tags'], dict):
            tags = self._config['tags']

        if not tags:
            return ""

        for name, value in six.iteritems(tags):
            if helpers.has_len(name) \
               and helpers.has_len(value):
                r.append((name, value))

        return urisup.unsplit_query(r)

    def set_options(self, options, from_boto = False):
        if not options or not isinstance(options, dict):
            options = {}

        if from_boto and self._client_cfg:
            for opt in _CONFIG_OPTIONS:
                options[opt] = getattr(self._client_cfg, opt)

            # pylint: disable-msg=no-member
            if self._client_cfg.retries and 'max_attempts' in self._client_cfg.retries:
                options['max_retries'] = self._client_cfg.retries['max_attempts']

        self.parse_headers = bool(options.get('parse_headers'))

        if 'max_redirs' not in options:
            options['max_redirs'] = DEFAULT_MAX_REDIRS

        if not options.get('max_retries'):
            options['max_retries'] = DEFAULT_MAX_RETRIES

        if not options.get('retry_delay'):
            options['retry_delay'] = DEFAULT_RETRY_DELAY

        if options['retry_delay'] < 0:
            options['retry_delay'] = 0

        for x in ('connect', 'read'):
            opt = "%s_timeout" % x
            if opt in options:
                setattr(self, opt, int(options[opt]))
            else:
                setattr(self, opt, int(DEFAULT_TIMEOUT))

        self.max_redirs   = int(options['max_redirs'])
        self.max_retries  = int(options['max_retries'])
        self.retry_delay  = float(options['retry_delay'])
        self.retry_factor = bool(options.get('retry_factor'))


class XtrargonTxProgress(object): # pylint: disable-msg=useless-object-inheritance
    def __init__(self, filename, delete = False, killed = False):
        self._filename    = filename
        self._delete      = delete
        self._killed      = killed
        self._size        = float(os.path.getsize(filename))
        self._seen_so_far = 0
        self._lock        = threading.Lock()

    def __call__(self, bytes_amount):
        if self._killed:
            raise KeyboardInterrupt()

        with self._lock:
            self._seen_so_far += bytes_amount
            if self._size:
                percentage = (self._seen_so_far / self._size) * 100
            else:
                percentage = 100

            if not int(percentage) % 25:
                LOG.debug("XtrargonTxProgress: %r  %s / %s  (%.2f%%)",
                          self._filename,
                          bitmath.Byte(self._seen_so_far).best_prefix().format('{value:.2f} {unit}'),
                          bitmath.Byte(self._size).best_prefix().format('{value:.2f} {unit}'),
                          percentage)

            if self._delete and percentage >= 100:
                try:
                    os.unlink(self._filename)
                except Exception:
                    pass


class XtrargonS3Command(object): # pylint: disable-msg=useless-object-inheritance
    def __init__(self, client = None, config = None, killed = False):
        if not isinstance(config, XtrargonS3Helper):
            self._config = XtrargonS3Helper()
        else:
            self._config = config

        if client:
            self._client = client
        else:
            self._client = self._config.botocore_client(event_hooks = XtrargonBotoNullEventHooks())

        self._pc          = None
        self._raw_headers = None
        self._request     = None
        self._req_dict    = None
        self._killed      = killed

    def kill(self):
        self._killed = True

    def killed(self):
        return self._killed

    def _prepare(self, op, params = None, body = None):
        if params is None:
            params = {}

        op_model = self._client._service_model.operation_model(op) # pylint: disable-msg=protected-access

        request_context = {
            'client_region': self._client.meta.region_name,
            'client_config': self._client.meta.config,
            'has_streaming_input': op_model.has_streaming_input,
            'auth_type': op_model.auth_type,
        }

        # pylint: disable-msg=protected-access
        self._req_dict = self._client._convert_to_request_dict(
            params,
            op_model,
            context=request_context)

        if body is not None:
            self._req_dict['body'] = body

        self._request = botocore.awsrequest.create_request_object(self._req_dict)
        self._client._request_signer.sign(op, self._request) # pylint: disable-msg=protected-access

        # pylint: disable-msg=c-extension-no-member
        self._pc.setopt(pycurl.URL, str(self._request.url))
        self._pc.setopt(pycurl.CONNECTTIMEOUT, self._config.connect_timeout)
        self._pc.setopt(pycurl.CUSTOMREQUEST, str(self._request.method))
        self._pc.setopt(pycurl.HEADER, 1)
        self._pc.setopt(pycurl.FOLLOWLOCATION, 1)
        self._pc.setopt(pycurl.MAXREDIRS, self._config.max_redirs)
        self._pc.setopt(pycurl.LOW_SPEED_TIME, 5)
        self._pc.setopt(pycurl.LOW_SPEED_LIMIT, 1024)
        self._pc.setopt(pycurl.NOPROGRESS, 1)
        self._pc.setopt(pycurl.NOSIGNAL, 1)
        self._pc.setopt(pycurl.TIMEOUT, self._config.read_timeout)
        self._pc.setopt(pycurl.VERBOSE, 0)
        self._pc.setopt(pycurl.WRITEFUNCTION, lambda x: None)

        if self._request.method == 'HEAD':
            self._pc.setopt(pycurl.NOBODY, 1)
        elif self._req_dict['body']:
            self._pc.setopt(pycurl.POSTFIELDS, self._req_dict['body'])

        headers = list(DEFAULT_HTTP_HEADERS)

        for x in self._request.headers.items():
            headers.append(str(": ".join(x)))

        self._pc.setopt(pycurl.HTTPHEADER, headers)

    def _parse_headers(self):
        if self._raw_headers:
            self._raw_headers.seek(0)
            self._raw_headers.readline()
            return Parser().parse(self._raw_headers, True).as_string()[:-1]

        return None

    def _clean(self, pc = False):
        if self._raw_headers:
            self._raw_headers.close()

        if pc and self._pc:
            self._pc.close()

    def run(self, op, params = None, body = None):
        r                 = {'headers':    None,
                             'code':       None,
                             'nb_retries': 0}
        if self._killed:
            return r

        error             = {'Error': {'Code': None,
                                       'Message': ''},
                             'ResponseMetadata': {}}

        self._pc          = None
        self._raw_headers = None
        self._request     = None
        self._req_dict    = None
        retry_delay       = self._config.retry_delay

        try:
            self._pc = pycurl.Curl() # pylint: disable-msg=c-extension-no-member
            self._prepare(op, params, body)

            while not self._killed:
                try:
                    if self._config.parse_headers:
                        self._raw_headers = StringIO()
                        self._pc.setopt(pycurl.HEADERFUNCTION, self._raw_headers.write) # pylint: disable-msg=c-extension-no-member

                    self._pc.perform()
                    r['code'] = self._pc.getinfo(self._pc.HTTP_CODE)

                    if self._config.parse_headers:
                        r['headers'] = self._parse_headers()

                    if r['code'] < 400:
                        break
                    error['Error']['Code'] = str(r['code'])
                    LOG.error("return code: %r", r['code'])
                    raise botocore.exceptions.ClientError(error, op)
                except Exception as e:
                    LOG.warning("%r", e)
                    r['nb_retries'] += 1
                    self._clean()
                    if r['nb_retries'] >= self._config.max_retries:
                        raise
                    if self._config.retry_factor:
                        retry_delay = (self._config.retry_delay
                                       * (2 * (r['nb_retries'] - 1)))
                    if retry_delay:
                        time.sleep(retry_delay)
        finally:
            self._request  = None
            self._req_dict = None
            self._clean(pc = True)

        return r

    def __call__(self, op, params = None, body = None):
        return self.run(op, params, body)
