# -*- coding: utf-8 -*-
# Copyright (C) 2019 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.modules.backups"""

import copy
import logging
import os
import threading
import time

from datetime import datetime

import six

from dwho.adapters.redis import DWhoAdapterRedis
from dwho.classes.errors import DWhoConfigurationError
from dwho.classes.modules import MODULES
from dwho.classes.inotify import DWhoInotifyConfig
from dwho.classes.notifiers import DWhoPushNotifications
from httpdis.ext.httpdis_json import HttpReqErrJson
from sonicprobe import helpers

from sonicprobe.libs.moresynchro import RWLock

from ..classes.abstract.modules import XtrargonDBModuleBase
from ..classes.common import (DATE_FORMAT,
                              STATUS_FAILURE,
                              STATUS_SUCCESS,
                              STATUS_STOPPED,
                              STATUS_TRIGGERED,
                              STATUS_UNKNOWN)
from ..classes.scan_paths import BackupsScanPaths
from ..classes.s3 import XtrargonS3Helper


LOG = logging.getLogger('xtrargon_ng.modules.backups')


class BackupsModule(XtrargonDBModuleBase):
    MODULE_NAME = 'backups'

    LOCK = RWLock()

    @staticmethod
    def _load_path_params(toppath, params, ino_params = None):
        ino_path_params = {}

        if isinstance(ino_params, dict):
            ino_path_params = copy.deepcopy(ino_params)

        r = {'compress_dir': None,
             'dest_params': None,
             'exclude_patterns': None,
             'follow': True,
             'inotify': {},
             'max_attempts': 1,
             'second_phase': {}}

        if not params or not isinstance(params, dict):
            return r

        r['follow'] = bool(params.get('follow', True))
        r['max_attempts'] = int(params.get('max_attempts') or 1)
        if r['max_attempts'] < 1:
            r['max_attempts'] = 1

        if 'exclude_files' in params:
            r['exclude_patterns'] = DWhoInotifyConfig.load_exclude_patterns(params['exclude_files'])
            ino_path_params['exclude_files'] = params['exclude_files']

        for x in ('compress_dir', 'dest_params', 'inotify', 'second_phase'):
            if isinstance(params.get(x), dict):
                r[x] = copy.deepcopy(params[x])

        if r['second_phase'].get('enabled', True) \
           and r['second_phase'].get('walk_from') == 'inotify':
            ino_path_params.update(r['inotify'])
            if not isinstance(ino_path_params.get('plugins'), dict):
                ino_path_params['plugins'] = {}
            r['inotify'] = {}
            r['inotify']['plugins'] = copy.deepcopy(ino_path_params['plugins'])
            r['inotify']['paths'] = {os.path.abspath(toppath): ino_path_params}
        else:
            r['inotify'] = {}
            r['inotify']['paths'] = {}
            r['inotify']['plugins'] = {}
            if r['second_phase'].get('walk_from') == 'inotify':
                r['second_phase']['walk_from'] = None

        return r

    # pylint: disable-msg=attribute-defined-outside-init
    def _chk_config(self):
        if not self.config.get('backups'):
            raise DWhoConfigurationError("missing backups configuration")

        if not self.config['backups'].get('trigger_file'):
            self.is_triggerable = False
        else:
            self.is_triggerable = helpers.has_len(self.config['backups']['trigger_file'])

        for x in ('db', 'inotify'):
            if not isinstance(self.config['backups'].get(x), dict):
                self.config['backups'][x] = {}

        if isinstance(self.config.get('db'), dict) \
           and not self.config['backups']['db']:
            self.config['backups']['db'] = copy.deepcopy(self.config['db'])

        ino_params = {}

        if isinstance(self.config['backups']['inotify'].get('events'), list):
            ino_params['events'] = self.config['backups']['inotify'].pop('events', None)

        if isinstance(self.config['backups']['inotify'].get('plugins'), dict):
            ino_params['plugins'] = self.config['backups']['inotify'].pop('plugins', None)

        if 'alive' not in self.config['backups']['db']:
            self.config['backups']['db']['alive'] = True

        for x in ('paths', 's3'):
            if x not in self.config['backups'] \
             or not isinstance(self.config['backups'][x], dict):
                raise DWhoConfigurationError("missing backups %s" % x)

        cfg_paths = copy.deepcopy(self.config['backups']['paths'])

        for toppath, params in six.iteritems(cfg_paths):
            if not params.get('enabled', True):
                del self.config['backups']['paths'][toppath]
            else:
                self.config['backups']['paths'][toppath] = self._load_path_params(toppath, params, ino_params)

    def _load_s3helper(self):
        r = XtrargonS3Helper(self.config['backups']['s3'])
        r.load_credentials()

        if not r.get_bucket():
            raise DWhoConfigurationError("missing default bucket setting")

        return r

    def _check_redis_server(self, prefix):
        redisconn = DWhoAdapterRedis(self.config, prefix = prefix)
        if not redisconn.servers:
            raise DWhoConfigurationError("missing redis server for %s" % prefix)

        redisconn.ping()

        return redisconn

    def _unlink_trigger_file(self):
        if self.is_triggerable \
           and os.path.isfile(self.config['backups']['trigger_file']):
            os.unlink(self.config['backups']['trigger_file'])

    # pylint: disable-msg=attribute-defined-outside-init
    def safe_init(self, options):
        self.lock_timeout   = self.config['general']['lock_timeout']

        self.is_running     = threading.Event()
        self.is_scanning    = threading.Event()
        self.is_triggerable = False

        self._bkp_scan      = None
        self._errors        = []
        self._state         = {}
        self.reset_state()

        self._chk_config()

        if self.config['backups']['db']['alive']:
            self._init_db()

        self._bkp_history   = self._check_redis_server('backups_history')
        self.file_sync      = self._check_redis_server('file_sync')

        self.s3helper       = self._load_s3helper()

        self.notifiers      = None
        if self.config['general'].get('notifiers_path'):
            self.notifiers = DWhoPushNotifications(self.server_id,
                                                   self.config['general']['notifiers_path'])

    def _terminate_workers(self):
        if self._bkp_scan:
            self._bkp_scan.terminate()
            self._bkp_scan = None

        self._terminate_lock_db()

    def _stop(self):
        self._unlink_trigger_file()
        self._terminate_workers()

    def at_stop(self):
        self._stop()

    def _start(self, start = True, clear_events = True):
        self._terminate_workers()
        self.reset_state()

        self._init_lock_db(self.config['backups']['db'])

        self._bkp_scan = BackupsScanPaths(self)
        self._bkp_scan.daemon = True

        if clear_events:
            self.is_running.clear()
            self.is_scanning.clear()
            self._clear_lock_db()

        if start:
            self._bkp_scan.start()
            self._start_lock_db()
            self.is_running.set()

    def add_errors(self, errors):
        if isinstance(errors, six.string_types):
            errors = [errors]

        if not isinstance(errors, list):
            return

        for error in errors:
            if error not in self._errors:
                self._errors.append(error)

    def has_errors(self):
        return len(self._errors) > 0

    def get_state(self, status = None, end = False):
        if 'status' not in self._state:
            self.reset_state()

        self._state['errors'] = self._errors

        if status:
            self._state['status'] = status
        elif self._state['status'] == STATUS_UNKNOWN \
           and self.is_triggerable \
           and os.path.isfile(self.config['backups']['trigger_file']):
            self._state['status'] = STATUS_TRIGGERED
        elif self._state['status'] == STATUS_TRIGGERED \
           and self.is_triggerable \
           and not os.path.isfile(self.config['backups']['trigger_file']):
            self._state['status'] = STATUS_UNKNOWN

        started_at = self._state.get('started_at')
        if started_at:
            started_at = datetime.strptime(started_at, DATE_FORMAT)
            duration   = (datetime.now() - started_at).total_seconds()
            self._state['duration'] = duration

        if not end or self._state['ended_at']:
            return self._state

        self._state['ended_at'] = time.strftime(DATE_FORMAT)

        if self._state['status'] == STATUS_STOPPED:
            self._state['errors'].append('stopped before completing')
        elif not self.has_errors():
            self._state['status'] = STATUS_SUCCESS
        else:
            self._state['status'] = STATUS_FAILURE

        return self._state

    def reset_state(self):
        self._errors = []
        self._state  = {'status':     STATUS_UNKNOWN,
                        'started_at': None,
                        'ended_at':   None,
                        'uuid':       None,
                        'duration':   -1,
                        'server_id':  self.server_id,
                        'errors':     []}

    def get_state_value(self, key):
        return self._state.get(key)

    def set_state_value(self, key, value):
        self._state[key] = value

    def start(self, request): # pylint: disable-msg=unused-argument
        if self.is_running.is_set():
            raise HttpReqErrJson(409, "backup is already processing")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            self._start()
            self.is_scanning.wait(self.lock_timeout)
            return self.get_state()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def status(self, request): # pylint: disable-msg=unused-argument
        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            return self.get_state()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def stop(self, request): # pylint: disable-msg=unused-argument
        self._unlink_trigger_file()

        if not self.is_running.is_set():
            raise HttpReqErrJson(409, "backup isn't running")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            self._stop()
            xtime = time.time()

            while self.is_running.is_set():
                if time.time() - xtime > self.lock_timeout:
                    self.add_errors("unable to stop backup")
                    break
                time.sleep(0.1)
            return self.get_state(end = True)
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def trigger(self, request): # pylint: disable-msg=unused-argument
        if self.is_running.is_set():
            raise HttpReqErrJson(409, "backup is already processing")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            if not self.config['backups'].get('trigger_file'):
                return False

            if not self.is_triggerable:
                raise HttpReqErrJson(500, "invalid option trigger_file in section backups")

            if os.path.isfile(self.config['backups']['trigger_file']):
                return STATUS_TRIGGERED

            helpers.touch(self.config['backups']['trigger_file'])
            self.set_state_value('status', STATUS_TRIGGERED)

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(BackupsModule())
    _start()
